﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Microsoft.Phone.Controls;
using Microsoft.Devices.Sensors;

namespace ViewHorizontalCoordinates
{
    public partial class MainPage : PhoneApplicationPage
    {
        double PIXELS_PER_DEGREE = 15;
        Motion motion;
        HorizontalCoordinateProjection coordinateProjection = new HorizontalCoordinateProjection();
        Brush lineBrush = new SolidColorBrush(System.Windows.Media.Color.FromArgb(255, 128, 128, 255));
        Brush textBrush = new SolidColorBrush(System.Windows.Media.Color.FromArgb(255, 192, 192, 255));
        double width, height;
        string[] compass = { "North", "NE", "East", "SE", "South", "SW", "West", "NW" };
        bool elementsCreated;

        public MainPage()
        {
            InitializeComponent();

            if (Motion.IsSupported)
            {
                motion = new Motion();
                motion.TimeBetweenUpdates = TimeSpan.FromMilliseconds(100);
                motion.CurrentValueChanged += OnMotionCurrentValueChanged;
            }

            Loaded += (sender, args) =>
            {
                width = ContentPanel.ActualWidth;
                height = ContentPanel.ActualHeight;
            };
        }

        protected override void OnNavigatedTo(NavigationEventArgs args)
        {
            if (motion != null)
            {
                try { motion.Start(); }
                catch { }
            }
            base.OnNavigatedTo(args);
        }

        protected override void OnNavigatedFrom(NavigationEventArgs args)
        {
            if (motion != null)
            {
                motion.Stop();
            }
            base.OnNavigatedFrom(args);
        }

        void OnMotionCurrentValueChanged(object sender, SensorReadingEventArgs<MotionReading> args)
        {
            this.Dispatcher.BeginInvoke(() =>
            {
                Microsoft.Xna.Framework.Matrix matrix = args.SensorReading.Attitude.RotationMatrix;
                HorizontalCoordinate viewCenter = HorizontalCoordinate.FromMotionMatrix(matrix);
                coordinateProjection.SetViewCenter(viewCenter);

                // Rotate the canvas
                rotate.Angle = -viewCenter.Tilt;

                int childIndex = 0;

                // Draw azimuth lines
                for (double azimuth = 0; azimuth < 360; azimuth += 15)
                {
                    for (double altitude = -90; altitude < 90; altitude += 15)
                    {
                        HorizontalCoordinate coord1 = new HorizontalCoordinate(azimuth, altitude);
                        HorizontalCoordinate coord2 = new HorizontalCoordinate(azimuth, altitude + 15);
                        DrawLine(ref childIndex, coord1, coord2);
                    }
                }

                // Draw altitude lines
                for (double altitude = -75; altitude < 90; altitude += 15)
                {
                    for (double azimuth = 0; azimuth < 360; azimuth += 15)
                    {
                        HorizontalCoordinate coord1 = new HorizontalCoordinate(azimuth, altitude);
                        HorizontalCoordinate coord2 = new HorizontalCoordinate(azimuth + 15, altitude);
                        DrawLine(ref childIndex, coord1, coord2);
                    }
                }

                // Draw text for altitude angles
                for (double altitude = -75; altitude < 90; altitude += 15)
                {
                    HorizontalCoordinate coord = new HorizontalCoordinate(viewCenter.Azimuth, altitude);
                    DrawText(ref childIndex, altitude.ToString(), coord);
                }

                // Draw text for azimuth compass points
                for (double azimuth = 0; azimuth < 360; azimuth += 45)
                {
                    double altitude = Math.Min(80, Math.Max(-80, viewCenter.Altitude));
                    HorizontalCoordinate coord = new HorizontalCoordinate(azimuth, altitude);
                    DrawText(ref childIndex, compass[(int)(azimuth / 45)], coord);
                }
                elementsCreated = true;
            });
        }

        void DrawLine(ref int childIndex, HorizontalCoordinate coord1, HorizontalCoordinate coord2)
        {
            Point point1 = CalculateScreenPoint(coord1);
            Point point2 = CalculateScreenPoint(coord2);

            // If points are out of range, move them offscreen
            if (double.IsNaN(point1.X) || double.IsNaN(point2.X))
            {
                point1 = point2 = new Point(-1000, -1000);
            }

            Line line = null;

            // Create Line first time through
            if (!elementsCreated)
            {
                line = new Line
                {
                    Stroke = lineBrush,
                    StrokeThickness = 3
                };
                canvas.Children.Add(line);
            }
            // Get created Line otherwise
            else
            {
                line = canvas.Children[childIndex] as Line;
                childIndex += 1;
            }

            line.X1 = point1.X;
            line.Y1 = point1.Y;
            line.X2 = point2.X;
            line.Y2 = point2.Y;
        }

        void DrawText(ref int childIndex, string str, HorizontalCoordinate coord)
        {
            Point point = CalculateScreenPoint(coord);

            // If point is out of range, move them offscreen
            if (double.IsNaN(point.X))
            {
                point = new Point(-1000, -1000);
            }

            TextBlock txtblk = null;

            // Create TextBlock first time through
            if (!elementsCreated)
            {
                txtblk = new TextBlock
                {
                    Text = str,
                    FontSize = 48,
                    Foreground = textBrush,
                };
                canvas.Children.Add(txtblk);
            }
            // Get created TextBlock otherwise
            else
            {
                txtblk = canvas.Children[childIndex] as TextBlock;
                childIndex += 1;
            }

            Canvas.SetLeft(txtblk, point.X - txtblk.ActualWidth / 2);
            Canvas.SetTop(txtblk, point.Y - txtblk.ActualHeight / 2);
        }

        Point CalculateScreenPoint(HorizontalCoordinate horizontalCoordinate)
        {
            double horzAngle, vertAngle;
            coordinateProjection.GetAngleOffsets(horizontalCoordinate, out horzAngle, out vertAngle);

            // Use NaN to indicate points clearly out of range of the screen
            float x = float.NaN;
            float y = float.NaN;

            if (horzAngle > -90 && horzAngle < 90 && vertAngle > -90 && vertAngle < 90)
            {
                x = (float)(width / 2 + PIXELS_PER_DEGREE * horzAngle);
                y = (float)(height / 2 + PIXELS_PER_DEGREE * vertAngle);
            }
            return new Point(x, y);
        }
    }
}