using System;
using System.Device.Location;
using Microsoft.Devices.Sensors;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;
using Petzold.Astronomy;
using Petzold.Phone.Xna;

namespace ViewThreeCoordinates
{
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        const int WIDTH = 480;
        const int HEIGHT = 800;
        const int PIXELS_PER_DEGREE = 15;
        static readonly Angle DIVISION_ANGLE = Angle.FromDegrees(15);
        static readonly int numAzimuthAngles = (int)(Angle.Circle / DIVISION_ANGLE);
        static readonly int numPolarAngles = (int)(Angle.Straight / DIVISION_ANGLE) + 1;

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        // Graphics relatied
        SpriteFont labelFont;
        Matrix displayMatrix = Matrix.Identity;
        LineRenderer lineRenderer;
        HorizontalCoordinateProjection coordinateProjection = new HorizontalCoordinateProjection();

        // Sensors
        Motion motion;
        GeoCoordinateWatcher geoWatcher;

        // Displayed data
        Vector2[,] horizontalScreenPoints = new Vector2[numAzimuthAngles, numPolarAngles];
        Vector2[,] equatorialScreenPoints = new Vector2[numAzimuthAngles, numPolarAngles];
        Vector2[,] eclipticScreenPoints = new Vector2[numAzimuthAngles, numPolarAngles];

        class Label
        {
            public Label(string text)
            {
                this.Text = text;
            }

            public string Text;
            public Vector2 Size;
            public Vector2 Position;
        }

        Label[] compassLabels = { new Label("North"), new Label("NE"), new Label("East"), new Label("SE"),
                                  new Label("South"), new Label("SW"), new Label("West"), new Label("NW") };

        Label[] rightAscensionLabels = new Label[24];

        Label[] zodiacLabels = { new Label("0"), new Label("Aries"), new Label("30"), new Label("Taurus"), 
                                 new Label("60"), new Label("Gemini"), new Label("90"), new Label("Cancer"), 
                                 new Label("120"), new Label("Leo"), new Label("150"), new Label("Virgo"),
                                 new Label("180"), new Label("Libra"), new Label("210"), new Label("Scorpio"), 
                                 new Label("240"), new Label("Sagittarius"), new Label("270"), new Label("Capricorn"), 
                                 new Label("300"), new Label("Aquarius"), new Label("330"), new Label("Pisces") };

        Label[] altitudeLabels = new Label[numPolarAngles - 2];

        int visibilityMode = 6;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            // Frame rate is 30 fps by default for Windows Phone.
            TargetElapsedTime = TimeSpan.FromTicks(333333);

            // Extend battery life under lock.
            InactiveSleepTime = TimeSpan.FromSeconds(1);

            // Switch to portrait mode
            graphics.IsFullScreen = true;
            graphics.PreferredBackBufferWidth = WIDTH;
            graphics.PreferredBackBufferHeight = HEIGHT;

            // Create Motion object
            if (Motion.IsSupported)
            {
                motion = new Motion
                {
                    TimeBetweenUpdates = this.TargetElapsedTime
                };
            }

            // Create GPS watcher
            geoWatcher = new GeoCoordinateWatcher();
        }

        protected override void OnActivated(object sender, EventArgs args)
        {
            if (motion != null)
            {
                try { motion.Start(); }
                catch { }
            }

            geoWatcher.Start();
            base.OnActivated(sender, args);
        }

        protected override void OnDeactivated(object sender, EventArgs args)
        {
            if (motion != null)
                motion.Stop();

            geoWatcher.Stop();
            base.OnDeactivated(sender, args);
        }

        protected override void Initialize()
        {
            // Initialize some labels
            for (int index = 0; index < altitudeLabels.Length; index++)
                altitudeLabels[index] = new Label(((index + 1) * DIVISION_ANGLE + -Angle.Right).ToString());

            for (int index = 0; index < rightAscensionLabels.Length; index++)
            {
                double hour = 24.0 * index / rightAscensionLabels.Length;
                rightAscensionLabels[index] = new Label(hour.ToString() + " h"); 
            }

            // Enable taps 
            TouchPanel.EnabledGestures = GestureType.Tap;
            base.Initialize();
        }

        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // Load the font and obtain text sizes
            labelFont = this.Content.Load<SpriteFont>("LabelFont");

            foreach (Label label in compassLabels)
                label.Size = labelFont.MeasureString(label.Text);

            foreach (Label label in altitudeLabels)
                label.Size = labelFont.MeasureString(label.Text);

            foreach (Label label in rightAscensionLabels)
                label.Size = labelFont.MeasureString(label.Text);

            foreach (Label label in zodiacLabels)
                label.Size = labelFont.MeasureString(label.Text);

            // Create the line-rendering class
            lineRenderer = new LineRenderer(this.GraphicsDevice);
        }

        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // If screen tapped, change visibility mode
            while (TouchPanel.IsGestureAvailable)
                if (TouchPanel.ReadGesture().GestureType == GestureType.Tap)
                    visibilityMode = (visibilityMode += 1) % 7;

            if (motion != null && motion.IsDataValid)
            {
                // Create a HorizontalCoordinate from the 3D rotation matrix
                Matrix matrix = motion.CurrentValue.Attitude.RotationMatrix;
                HorizontalCoordinate viewCenter = HorizontalCoordinate.From(matrix);

                // Use tilt to globally rotate the graphics
                displayMatrix = Matrix.CreateTranslation(-WIDTH / 2, -HEIGHT / 2, 0) *
                Matrix.CreateRotationZ((float)-viewCenter.Tilt.Radians) *
                Matrix.CreateTranslation(WIDTH / 2, HEIGHT / 2, 0);

                // Set the view center to the HorizontalCoordinateProjection
                coordinateProjection.SetViewCenter(viewCenter);

                // Get the current time
                Time time = Time.GetNow();

                // Get the current location
                GeoCoordinate location = geoWatcher.Position.Location;
                GeographicCoordinate geographicCoordinate = new GeographicCoordinate(Angle.FromDegrees(-location.Longitude),
                                                                                     Angle.FromDegrees(location.Latitude));

                for (int azimuthIndex = 0; azimuthIndex < horizontalScreenPoints.GetLength(0); azimuthIndex += 1)
                {
                    for (int altitudeIndex = 0; altitudeIndex < horizontalScreenPoints.GetLength(1); altitudeIndex += 1)
                    {
                        // Calculate positions of horizontal coordinates
                        Angle azimuth = DIVISION_ANGLE * azimuthIndex;
                        Angle altitude = -Angle.Right + DIVISION_ANGLE * altitudeIndex;
                        HorizontalCoordinate horizontalCoordinate = new HorizontalCoordinate(azimuth, altitude);
                        Vector2 point = CalculateScreenPoint(horizontalCoordinate);
                        horizontalScreenPoints[azimuthIndex, altitudeIndex] = point;

                        // Calculate positions of equatorial coordinates
                        Angle rightAscension = DIVISION_ANGLE * azimuthIndex;
                        Angle declination = -Angle.Right + DIVISION_ANGLE * altitudeIndex;
                        EquatorialCoordinate equatorialCoordinate = new EquatorialCoordinate(rightAscension, declination);
                        horizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time);
                        point = CalculateScreenPoint(horizontalCoordinate);
                        equatorialScreenPoints[azimuthIndex, altitudeIndex] = point;

                        // Calculate positions of ecliptic coordinates
                        Angle longitude = DIVISION_ANGLE * azimuthIndex;
                        Angle latitude = -Angle.Right + DIVISION_ANGLE * altitudeIndex;
                        EclipticCoordinate eclipticCoordinate = new EclipticCoordinate(longitude, latitude, 0);
                        equatorialCoordinate = EquatorialCoordinate.From(eclipticCoordinate, time);
                        horizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time);
                        point = CalculateScreenPoint(horizontalCoordinate);
                        eclipticScreenPoints[azimuthIndex, altitudeIndex] = point;
                    }
                }

                // Calculate positions of compass points
                for (int index = 0; index < compassLabels.Length; index++)
                {
                    Label label = compassLabels[index];
                    Angle azimuth = index * Angle.Circle / compassLabels.Length;
                    Angle altitude = Angle.FromDegrees(Math.Min(80, Math.Max(-80, viewCenter.Altitude.Degrees)));
                    HorizontalCoordinate horizontalCoordinate = new HorizontalCoordinate(azimuth, altitude);
                    Vector2 point = CalculateScreenPoint(horizontalCoordinate);
                    label.Position = new Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2);
                }

                // Calculate positions of altitude labels
                for (int index = 0; index < altitudeLabels.Length; index++)
                {
                    Label label = altitudeLabels[index];
                    Angle altitude = (index + 1) * DIVISION_ANGLE - Angle.Right;
                    HorizontalCoordinate horizontalCoordinate = new HorizontalCoordinate(viewCenter.Azimuth, altitude);
                    Vector2 point = CalculateScreenPoint(horizontalCoordinate);
                    label.Position = new Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2);
                }

                // Calculate positions of right ascension labels
                for (int index = 0; index < rightAscensionLabels.Length; index++)
                {
                    Label label = rightAscensionLabels[index];
                    Angle rightAscension = index * Angle.Circle / rightAscensionLabels.Length;
                    EquatorialCoordinate equatorialCoordinate = new EquatorialCoordinate(rightAscension, Angle.Zero);
                    HorizontalCoordinate horizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time);
                    Vector2 point = CalculateScreenPoint(horizontalCoordinate);
                    label.Position = new Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2);
                }

                // Calculate positions of zodiac labels
                for (int index = 0; index < zodiacLabels.Length; index++)
                {
                    Label label = zodiacLabels[index];
                    Angle longitude = index * Angle.Circle / zodiacLabels.Length;
                    EclipticCoordinate eclipticCoordinate = new EclipticCoordinate(longitude, Angle.Zero, 0);
                    EquatorialCoordinate equatorialCoordinate = EquatorialCoordinate.From(eclipticCoordinate, time);
                    HorizontalCoordinate horizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time);
                    Vector2 point = CalculateScreenPoint(horizontalCoordinate);
                    label.Position = new Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2);
                }
            }
            base.Update(gameTime);
        }

        Vector2 CalculateScreenPoint(HorizontalCoordinate horizontalCoordinate)
        {
            Angle horzAngle, vertAngle;
            coordinateProjection.GetAngleOffsets(horizontalCoordinate, out horzAngle, out vertAngle);

            // Use NaN to indicate points clearly out of range of the screen
            float x = float.NaN;
            float y = float.NaN;

            if (horzAngle > -Angle.Right && horzAngle < Angle.Right &&
                vertAngle > -Angle.Right && vertAngle < Angle.Right)
            {
                x = (float)(WIDTH / 2 + PIXELS_PER_DEGREE * horzAngle.Degrees);
                y = (float)(HEIGHT / 2 + PIXELS_PER_DEGREE * vertAngle.Degrees);
            }
            return new Vector2(x, y);
        }

        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(new Color(0, 0, 32));

            spriteBatch.Begin(SpriteSortMode.Immediate, null, null, null, null, null, displayMatrix);

            if (((visibilityMode + 1) & 1) != 0)
                DrawCoordinates(horizontalScreenPoints, compassLabels, altitudeLabels, Color.Green);

            if (((visibilityMode + 1) & 2) != 0)
                DrawCoordinates(equatorialScreenPoints, rightAscensionLabels, null, Color.Blue);

            if (((visibilityMode + 1) & 4) != 0)
                DrawCoordinates(eclipticScreenPoints, zodiacLabels, null, Color.Red);

            spriteBatch.End();
            base.Draw(gameTime);
        }

        void DrawCoordinates(Vector2[,] screenPositions, Label[] azimuthLabels, Label[] polarAngleLabels, Color color)
        {
            int numAzimuths = screenPositions.GetLength(0);
            int numAltitudes = screenPositions.GetLength(1);

            // Draw circles of altitude
            for (int altitudeIndex = 1; altitudeIndex < numAltitudes - 1; altitudeIndex += 1)
            {
                float thickness = altitudeIndex == numAltitudes / 2 ? 5 : 2;

                for (int azimuthIndex = 0; azimuthIndex < numAzimuths; azimuthIndex += 1)
                {
                    Vector2 point1 = screenPositions[azimuthIndex, altitudeIndex];
                    Vector2 point2 = screenPositions[(azimuthIndex + 1) % numAzimuths, altitudeIndex];

                    if (!float.IsNaN(point1.X) && !float.IsNaN(point2.X))
                        lineRenderer.Draw(spriteBatch, point1, point2, thickness, color);
                }
            }

            // Draw azimuth lines
            for (int azimuthIndex = 0; azimuthIndex < numAzimuths; azimuthIndex += 1)
            {
                for (int altitudeIndex = 0; altitudeIndex < numAltitudes - 1; altitudeIndex += 1)
                {
                    Vector2 point1 = screenPositions[azimuthIndex, altitudeIndex];
                    Vector2 point2 = screenPositions[azimuthIndex, altitudeIndex + 1];

                    if (!float.IsNaN(point1.X) && !float.IsNaN(point2.X))
                        lineRenderer.Draw(spriteBatch, point1, point2, 2, color);
                }
            }

            // Draw azimuth labels
            if (azimuthLabels != null)
                foreach (Label label in azimuthLabels)
                    if (!float.IsNaN(label.Position.X))
                        spriteBatch.DrawString(labelFont, label.Text, label.Position, color);

            // Draw polar angle labels
            if (polarAngleLabels != null)
                foreach (Label label in polarAngleLabels)
                    if (!float.IsNaN(label.Position.X))
                        spriteBatch.DrawString(labelFont, label.Text, label.Position, color);
        }
    }
}
